import time
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from alitra import Position
from pydantic import BaseModel, Field

from isar.apis.models.models import InputPose, InputPosition
from isar.config.settings import settings
from isar.mission_planner.mission_planner_interface import MissionPlannerError
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.step import (
    DockingProcedure,
    DriveToPose,
    Localize,
    RecordAudio,
    STEPS,
    TakeImage,
    TakeThermalImage,
    TakeThermalVideo,
    TakeVideo,
)
from robot_interface.models.mission.task import Task


class InspectionTypes(str, Enum):
    image: str = "Image"
    thermal_image: str = "ThermalImage"
    video: str = "Video"
    thermal_video: str = "ThermalVideo"
    audio: str = "Audio"


class TaskType(str, Enum):
    Inspection: str = "inspection"
    DriveTo: str = "drive_to"
    Localization: str = "localization"
    Dock: str = "dock"


class StartMissionInspectionDefinition(BaseModel):
    type: InspectionTypes = Field(default=InspectionTypes.image)
    inspection_target: InputPosition
    analysis_type: Optional[str] = None
    duration: Optional[float] = None
    metadata: Optional[dict] = None
    id: Optional[str] = None


class StartMissionTaskDefinition(BaseModel):
    type: TaskType = Field(default=TaskType.Inspection)
    pose: InputPose
    inspections: List[StartMissionInspectionDefinition]
    tag: Optional[str] = None
    id: Optional[str] = None


class StartMissionDefinition(BaseModel):
    tasks: List[StartMissionTaskDefinition]
    id: Optional[str] = None
    name: Optional[str] = None


def to_isar_mission(mission_definition: StartMissionDefinition) -> Mission:
    isar_tasks: List[Task] = []
    all_steps_in_mission: List[STEPS] = []

    for task in mission_definition.tasks:
        steps: List[STEPS] = generate_steps(task)
        all_steps_in_mission.extend(steps)

        isar_task: Task = Task(steps=steps, tag_id=task.tag)
        if task.id:
            isar_task.id = task.id
        isar_tasks.append(isar_task)

    if not isar_tasks:
        raise MissionPlannerError("Mission does not contain any valid tasks")

    check_for_duplicate_ids(isar_tasks)
    check_for_duplicate_ids(all_steps_in_mission)

    isar_mission: Mission = Mission(tasks=isar_tasks)

    if mission_definition.name:
        isar_mission.name = mission_definition.name
    else:
        isar_mission.name = _build_mission_name()

    if mission_definition.id:
        isar_mission.id = mission_definition.id

    return isar_mission


def check_for_duplicate_ids(items: Union[List[Task], List[STEPS]]):
    duplicate_ids = get_duplicate_ids(items=items)
    if len(duplicate_ids) > 0:
        raise MissionPlannerError(
            f"Failed to create as there were duplicate IDs which is not allowed "
            f"({duplicate_ids})"
        )


def generate_steps(task) -> List[STEPS]:
    steps: List[STEPS] = []
    try:
        match task.type:
            case TaskType.Inspection:
                steps.extend(generate_steps_for_inspection_task(task=task))
            case TaskType.DriveTo:
                steps.append(generate_steps_for_drive_to_task(task=task))
            case TaskType.Localization:
                steps.append(generate_steps_for_localization_task(task=task))
            case TaskType.Dock:
                steps.append(generate_steps_for_dock_task())
    except ValueError as e:
        raise MissionPlannerError(f"Failed to create task: {str(e)}")

    return steps


def generate_steps_for_inspection_task(task: StartMissionTaskDefinition) -> List[STEPS]:
    drive_step: DriveToPose = DriveToPose(pose=task.pose.to_alitra_pose())

    inspection_steps: List[STEPS] = [
        create_inspection_step(
            inspection_type=inspection.type,
            duration=inspection.duration,
            target=inspection.inspection_target.to_alitra_position(),
            tag_id=task.tag,
            metadata=inspection.metadata,
            id=inspection.id,
        )
        for inspection in task.inspections
    ]

    return [drive_step, *inspection_steps]


def generate_steps_for_drive_to_task(task: StartMissionTaskDefinition) -> DriveToPose:
    return DriveToPose(pose=task.pose.to_alitra_pose())


def generate_steps_for_localization_task(task: StartMissionTaskDefinition) -> Localize:
    return Localize(localization_pose=task.pose.to_alitra_pose())


def generate_steps_for_dock_task() -> DockingProcedure:
    return DockingProcedure(behavior="dock")


def create_inspection_step(
    inspection_type: InspectionTypes,
    duration: float,
    target: Position,
    tag_id: Optional[str],
    metadata: Optional[dict],
    id: Optional[str],
) -> STEPS:
    inspection_step_dict: Dict[str, Any] = {
        InspectionTypes.image.value: TakeImage(target=target),
        InspectionTypes.video.value: TakeVideo(target=target, duration=duration),
        InspectionTypes.thermal_image.value: TakeThermalImage(target=target),
        InspectionTypes.thermal_video.value: TakeThermalVideo(
            target=target, duration=duration
        ),
        InspectionTypes.audio.value: RecordAudio(target=target, duration=duration),
    }

    if inspection_type not in inspection_step_dict:
        raise ValueError(f"Inspection type '{inspection_type}' not supported")
    else:
        inspection_step = inspection_step_dict[inspection_type]

    if tag_id:
        inspection_step.tag_id = tag_id
    if metadata:
        inspection_step.metadata = metadata
    if id:
        inspection_step.id = id

    return inspection_step


def get_duplicate_ids(items: Union[List[Task], List[STEPS]]) -> List[str]:
    unique_ids: List[str] = []
    duplicate_ids: List[str] = []
    for item in items:
        id: str = item.id
        if id not in unique_ids:
            unique_ids.append(id)
        else:
            duplicate_ids.append(id)

    return duplicate_ids


def _build_mission_name() -> str:
    return f"{settings.PLANT_SHORT_NAME}{settings.ROBOT_NAME}{int(time.time())}"



from alitra import Frame, Orientation, Pose, Position

from isar.services.readers.base_reader import BaseReader
from robot_interface.models.mission.mission import Mission, Task
from robot_interface.models.mission.step import (
    DriveToPose,
    STEPS,
    TakeImage,
    TakeThermalImage,
)

drive_to_pose_1 = DriveToPose(
    pose=Pose(
        position=Position(x=2, y=2, z=0, frame=Frame(name="asset")),
        orientation=Orientation(
            x=0, y=0, z=0.7071068, w=0.7071068, frame=Frame(name="asset")
        ),
        frame=Frame(name="asset"),
    ),
)

take_image_1 = TakeImage(target=Position(x=2, y=3, z=1, frame=Frame(name="asset")))

drive_to_pose_2 = DriveToPose(
    pose=Pose(
        position=Position(x=4, y=4, z=0, frame=Frame(name="asset")),
        orientation=Orientation(
            x=0, y=0, z=-0.7071068, w=0.7071068, frame=Frame(name="asset")
        ),
        frame=Frame(name="asset"),
    ),
)

take_image_2 = TakeThermalImage(
    target=Position(x=4, y=3, z=1, frame=Frame(name="asset"))
)

drive_to_pose_3 = DriveToPose(
    pose=Pose(
        position=Position(x=0, y=0, z=0, frame=Frame(name="asset")),
        orientation=Orientation(x=0, y=0, z=0, w=1, frame=Frame(name="asset")),
        frame=Frame(name="asset"),
    ),
)

expected_mission = Mission(
    id="1",
    tasks=[
        Task(id="11", steps=[drive_to_pose_1, take_image_1]),
        Task(id="12", steps=[drive_to_pose_2, take_image_2]),
        Task(id="13", steps=[drive_to_pose_3]),
    ],
)


example_mission_dict = {
    "id": "1",
    "tasks": [
        {
            "steps": [
                {
                    "type": "drive_to_pose",
                    "pose": {
                        "position": {
                            "x": 2,
                            "y": 2,
                            "z": 0,
                            "frame": {"name": "asset"},
                        },
                        "orientation": {
                            "x": 0,
                            "y": 0,
                            "z": 0.7071068,
                            "w": 0.7071068,
                            "frame": {"name": "asset"},
                        },
                        "frame": {"name": "asset"},
                    },
                },
                {
                    "type": "take_image",
                    "target": {"x": 2, "y": 3, "z": 1, "frame": {"name": "asset"}},
                },
            ],
        },
        {
            "steps": [
                {
                    "type": "drive_to_pose",
                    "pose": {
                        "position": {
                            "x": 4,
                            "y": 4,
                            "z": 0,
                            "frame": {"name": "asset"},
                        },
                        "orientation": {
                            "x": 0,
                            "y": 0,
                            "z": -0.7071068,
                            "w": 0.7071068,
                            "frame": {"name": "asset"},
                        },
                        "frame": {"name": "asset"},
                    },
                },
                {
                    "type": "take_thermal_image",
                    "target": {"x": 4, "y": 3, "z": 1, "frame": {"name": "asset"}},
                },
            ],
        },
        {
            "steps": [
                {
                    "type": "drive_to_pose",
                    "pose": {
                        "position": {
                            "x": 0,
                            "y": 0,
                            "z": 0,
                            "frame": {"name": "asset"},
                        },
                        "orientation": {
                            "x": 0,
                            "y": 0,
                            "z": 0,
                            "w": 1,
                            "frame": {"name": "asset"},
                        },
                        "frame": {"name": "asset"},
                    },
                }
            ],
        },
    ],
}


def test_mission_definition() -> None:
    loaded_mission: Mission = BaseReader.dict_to_dataclass(
        dataclass_dict=example_mission_dict,
        target_dataclass=Mission,
        strict_config=True,
    )

    assert loaded_mission.id == expected_mission.id
    assert loaded_mission.status == expected_mission.status

    assert len(loaded_mission.tasks) == len(expected_mission.tasks)
    for i_task in range(len(loaded_mission.tasks)):
        loaded_task: Task = loaded_mission.tasks[i_task]
        expected_task: Task = expected_mission.tasks[i_task]

        assert loaded_task.status == expected_task.status
        assert loaded_task.tag_id == expected_task.tag_id

        for i_step in range(len(loaded_task.steps)):
            loaded_step: STEPS = loaded_task.steps[i_step]
            expected_step: STEPS = expected_task.steps[i_step]

            ignore_attributes = set(
                ("id", "status", "tag_id", "error_message", "analysis")
            )
            loaded_attributes = set(loaded_step.__dict__.keys()) - ignore_attributes
            expected_attributes = set(expected_step.__dict__.keys()) - ignore_attributes

            assert loaded_attributes == expected_attributes

            for attr in loaded_attributes:
                assert loaded_step.__dict__[attr] == expected_step.__dict__[attr]



import json
import logging
import time
from datetime import datetime
from logging import Logger
from queue import Queue
from threading import Thread
from typing import Optional

from isar.config.settings import settings
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states_enum import States
from robot_interface.models.exceptions.robot_exceptions import (
    RobotAPIException,
    RobotCommunicationException,
    RobotException,
)
from robot_interface.models.mission.status import RobotStatus
from robot_interface.robot_interface import RobotInterface
from robot_interface.telemetry.mqtt_client import MqttPublisher
from robot_interface.telemetry.payloads import RobotStatusPayload
from robot_interface.utilities.json_service import EnhancedJSONEncoder


class RobotStatusPublisher:
    def __init__(
        self, mqtt_queue: Queue, robot: RobotInterface, state_machine: StateMachine
    ):
        self.mqtt_publisher: MqttPublisher = MqttPublisher(mqtt_queue=mqtt_queue)
        self.robot: RobotInterface = robot
        self.state_machine: StateMachine = state_machine

    def _get_combined_robot_status(
        self, robot_status: RobotStatus, current_state: States
    ) -> RobotStatus:
        if robot_status == RobotStatus.Offline:
            return RobotStatus.Offline
        elif current_state == States.Idle and robot_status == RobotStatus.Available:
            return RobotStatus.Available
        elif robot_status == RobotStatus.Blocked:
            return RobotStatus.Blocked
        elif current_state != States.Idle or robot_status == RobotStatus.Busy:
            return RobotStatus.Busy
        return None

    def run(self) -> None:
        robot_status_monitor: RobotStatusMonitor = RobotStatusMonitor(robot=self.robot)
        robot_status_thread: Thread = Thread(
            target=robot_status_monitor.run,
            name="Robot Status Monitor",
            daemon=True,
        )
        robot_status_thread.start()

        previous_robot_status: Optional[RobotStatus] = None

        while True:
            time.sleep(settings.ROBOT_STATUS_PUBLISH_INTERVAL)

            combined_status: RobotStatus = self._get_combined_robot_status(
                robot_status=robot_status_monitor.robot_status,
                current_state=self.state_machine.current_state,
            )

            if previous_robot_status:
                if previous_robot_status == combined_status:
                    continue

            payload: RobotStatusPayload = RobotStatusPayload(
                isar_id=settings.ISAR_ID,
                robot_name=settings.ROBOT_NAME,
                robot_status=combined_status,
                previous_robot_status=previous_robot_status,
                current_isar_state=self.state_machine.current_state,
                current_mission_id=self.state_machine.current_mission.id
                if self.state_machine.current_mission
                else None,
                current_task_id=self.state_machine.current_task.id
                if self.state_machine.current_task
                else None,
                current_step_id=self.state_machine.current_step.id
                if self.state_machine.current_step
                else None,
                timestamp=datetime.utcnow(),
            )

            self.mqtt_publisher.publish(
                topic=settings.TOPIC_ISAR_ROBOT_STATUS,
                payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            )

            previous_robot_status = combined_status


class RobotStatusMonitor:
    def __init__(self, robot: RobotInterface):
        self.robot: RobotInterface = robot
        self.robot_status: RobotStatus = RobotStatus.Offline
        self.logger: Logger = logging.getLogger("robot_status_monitor")

    def run(self) -> None:
        while True:
            try:
                self.robot_status = self.robot.robot_status()
            except (
                RobotCommunicationException,
                RobotAPIException,
                RobotException,
            ) as e:
                self.logger.error(
                    f"Failed to get robot status because: {e.error_description}"
                )
            time.sleep(settings.ROBOT_API_STATUS_POLL_INTERVAL)
